"""
Some specialized list and array classes to store results obtained from R. These
classes provide means not to only access object items by index but also - sort
of like a dictionary - by key. However keys must not be unique or can even be
None. In those cases only the first item with that key is found.

Available classes:
- TaggedList
- TaggedArray
"""
import numpy


class TaggedList(object):
    # This code is mainly based on UserList.UserList and modified for tags
    """
    A tagged list is useful for additionally addressing individual items by
    name instead of only by index. In contrast to dictionaries multiple items
    can have the same name or key. However only the first one will be found.

    In many cases a TaggedList behaves like a normal list, however for lazyness
    reasons of the programmer not all methods are implemented yet.

    Example:
    l = TaggedList( [('v1', 1), ('v2', 2), 3, ('v2', 4)] )
    l[0]     # returns 1
    l['v1']  # returns 1
    l['v2']  # returns 2  (not 4 !)
    l[3]     # returns 4

    Data can be appended or inserted in the following way:
    l.insert(0, x=3)
    l['x']   # returns 3
    l[0]     # also returns 3

    l.append(y=3)
    l[-1]    # returns 3
    """
    def __init__(self, initlist=[]):
        """
        Items in initlist can either be
        - tuples of (key,values)
        - or plain values
        Keys can be None or empty strings in item tuples.
        """
        self.values = []
        self.keys = []
        for idx, item in enumerate(initlist):
            try:
                key, value = item
                key = None if key == '' else key
            except (TypeError, ValueError):
                value = item
                key = None
            finally:
                self.values.append(value)
                self.keys.append(key)

    def astuples(self):
        """
        Convert a TaggedList into a representation suitable to be provided
        to __init__()
        """
        return list(zip(self.keys, self.values))

    def __repr__(self):
        data = ["%s=%s" % (key, repr(value)) if key else "'%s'" % value
                for key, value in self.astuples()]
        return '<TaggedList(%s)>' % ', '.join(data)

    #    def __lt__(self, other): return self.values <  self.__cast(other)
    #    def __le__(self, other): return self.values <= self.__cast(other)
    #    def __eq__(self, other): return self.values == self.__cast(other)
    #    def __ne__(self, other): return self.values != self.__cast(other)
    #    def __gt__(self, other): return self.values >  self.__cast(other)
    #    def __ge__(self, other): return self.values >= self.__cast(other)
    #    def __cast(self, other):
    #        if isinstance(other, UserList): return other.data
    #        else: return other
    #    def __cmp__(self, other):
    #        return cmp(self.values, self.__cast(other))
    __hash__ = None  # Mutable sequence, so not hashable

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    def __contains__(self, item):
        return item in self.values

    def __len__(self):
        return len(self.values)

    def __getitem__(self, i):
        if type(i) == str:
            i = self.keys.index(i)
        return self.values[i]

    def __setitem__(self, i, item):
        if type(i) == str:
            i = self.keys.index[i]
        self.values[i] = item

    def __delitem__(self, i):
        if type(i) == str:
            i = self.keys.index[i]
            del self.keys[i]
        del self.values[i]

    def __getslice__(self, i, j):
        i = max(i, 0)
        j = max(j, 0)
        return self.__class__(self.astuples()[i:j])

    #    def __setslice__(self, i, j, other):
    #        i = max(i, 0); j = max(j, 0)
    #        if isinstance(other, UserList):
    #            self.values[i:j] = other.data
    #        elif isinstance(other, type(self.values)):
    #            self.values[i:j] = other
    #        else:
    #            self.values[i:j] = list(other)

    def __delslice__(self, i, j):
        raise NotImplementedError()
        # i = max(i, 0); j = max(j, 0)
        # del self.values[i:j]
        # del self.keys[i:j]

    def __add__(self, other):
        raise NotImplementedError()
        #        if isinstance(other, UserList):
        #            return self.__class__(self.values + other.data)
        #        elif isinstance(other, type(self.values)):
        #            return self.__class__(self.values + other)
        #        else:
        #            return self.__class__(self.values + list(other))

    def __radd__(self, other):
        raise NotImplementedError()
        #        if isinstance(other, UserList):
        #            return self.__class__(other.data + self.values)
        #        elif isinstance(other, type(self.values)):
        #            return self.__class__(other + self.values)
        #            return self.__class__(list(other) + self.values)

    def __iadd__(self, other):
        raise NotImplementedError()
        #        if isinstance(other, UserList):
        #            self.values += other.data
        #        elif isinstance(other, type(self.values)):
        #            self.values += other
        #        else:
        #            self.values += list(other)
        #        return self

    def __mul__(self, n):
        raise NotImplementedError()
        # return self.__class__(self.values*n)
    __rmul__ = __mul__

    def __imul__(self, n):
        raise NotImplementedError()
        # self.values *= n
        # return self

    def append(self, *value, **key_and_value):
        """
        Append an item to the list, either given as plain value or as a
        keyword-arg pair.
        Example:
            taggedlist.append(4)
        or
            taggedlist.append(k=4)
        """
        if len(value) == 1 and not key_and_value:
            key = None
            value = value[0]
        elif len(key_and_value) == 1 and not value:
            [(key, value)] = key_and_value.items()
        else:
            raise ValueError("Only either one single value or one single pair "
                             "of key/value is allowed")
        self.values.append(value)
        self.keys.append(key)

    def insert(self, i, *value, **key_and_value):
        """
        Insert an item in the list at position i, either given as plain value
        or as a keyword-arg pair.
        Example:
            taggedlist.insert(4, 'abc)
        or
            taggedlist.append(4, k='abc')
        """
        if len(value) == 1 and not key_and_value:
            key = None
            value = value[0]
        elif len(key_and_value) == 1 and not value:
            [(key, value)] = key_and_value.items()
        else:
            raise ValueError("Only either one single value or one single pair "
                             "of key/value is allowed")
        self.values.insert(i, value)
        self.keys.insert(i, key)

    def pop(self, i=-1):
        """
        Remove an item from the list. By default the last item will be removed.
        If an item at a specific position should be removed, pass an additional
        index arguemnt.
        """
        return self.values.pop(i)

    def remove(self, item):
        raise NotImplementedError()
        # self.values.remove(item)

    def count(self, item):
        return self.values.count(item)

    def index(self, item, *args):
        return self.values.index(item, *args)

    def reverse(self):
        self.values.reverse()
        self.keys.reverse()

    def sort(self, *args, **kwds):
        raise NotImplementedError()
        # self.values.sort(*args, **kwds)

    def extend(self, other):
        raise NotImplementedError()
        #  if isinstance(other, UserList):
        #      self.values.extend(other.data)
        #  else:
        #      self.values.extend(other)


class AttrArray(numpy.ndarray):
    """
    numpy.ndarray with additional "attr"-container.
    Used as base class for TaggedArray.
    """
    attr = None

    def __repr__(self):
        r = super(AttrArray, self).__repr__()
        if hasattr(self, 'attr'):
            return r[:-1] + ', attr=' + repr(self.attr) + ')'
        return r

    @classmethod
    def new(cls, data, attr):
        """
        Factory method to create AttrArray objects from ndarrays or Python
        lists.
        Usage:
            AttrArray.new(array([1, 2, 3, 4]), {'attr1': val1, 'attr2': val2})
        """
        if not isinstance(data, numpy.ndarray):
            # assume it is a Python list or any other valid data type
            # for arrays
            arr = numpy.array(data)
        else:
            arr = data

        attrArr = arr.view(cls)
        attrArr.attr = attr
        return attrArr


def asAttrArray(data, attr):
    return AttrArray.new(data, attr)


class TaggedArray(AttrArray):
    """
    A tagged array is useful for additionally addressing individual items by
    name instead of only by index. In contrast to dictionaries multiple items
    can have the same name or key. However only the first one will be found.

    In many cases a TaggedArray behaves like a normal array and is the
    equivalent for TaggedList.
    This class is basically only useful to translate results created by R into
    something useful in Python.

    Instances of TaggedArray should only be created using the factory function
    'asTaggedArray([values)], [tags])', where 'values' and 'tags' can be plain
    python lists or numpy-arrays.

    Example:
    l = asTaggedArray(array([1, 2, 3, 4]), ['v1', 'v2', 'v3', 'v4'])
    l[0]     # returns 1
    l['v1']  # returns 1
    l['v2']  # returns 2  (not 4 !)
    l[3]     # returns 4

    It is recommended not to do lots of manipulations that modify the
    structure of the arrary. This could lead to mismatched btw. tags and
    values (those are only very loosely coupled internally). However any type
    of mathematics like multiplying the array should be possible without
    problems.
    """
    attr = []

    def __repr__(self):
        r = super(AttrArray, self).__repr__()
        if hasattr(self, 'attr'):
            return r[:-1] + ', key=' + repr(self.attr) + ')'
        return r

    def __getitem__(self, idx_or_name):
        try:
            return numpy.ndarray.__getitem__(self, idx_or_name)
        except:
            pass
        try:
            return numpy.ndarray.__getitem__(self,
                                             self.attr.index(idx_or_name))
        except ValueError:
            raise KeyError('No key "%s" available for array' % idx_or_name)

    def keys(self):
        return self.attr[:]

    @classmethod
    def new(cls, data, tags):
        """
        Factory method to create TaggedArray objects from ndarrays or Python
        lists.
        Check the docs in TaggedArray for more information.
        Usage:
        l = TaggedArray.new(array([1, 2, 3, 4]), ['v1', 'v2', 'v3', 'v4'])
        l[0]     # returns 1
        l['v1']  # returns 1
        l['v2']  # returns 2  (not 4 !)
        l[3]     # returns 4
        """
        if len(tags) != len(data):
            raise ValueError('Number of keys must match size of array')
        if not isinstance(data, numpy.ndarray):
            # assume it is a Python list or any other valid data type
            # for arrays
            arr = numpy.array(data)
        else:
            arr = data

        taggedArr = arr.view(cls)
        taggedArr.attr = tags
        return taggedArr


def asTaggedArray(data, tags):
    return TaggedArray.new(data, tags)



"""
Exception classes for pyRserve
"""


class PyRserveError(Exception):
    pass


class REvalError(PyRserveError):
    """Indicates an error raised by R itself (not by Rserve)"""
    pass


class RConnectionRefused(PyRserveError):
    pass


class RResponseError(PyRserveError):
    pass


class RSerializationError(PyRserveError):
    pass


class PyRserveClosed(PyRserveError):
    pass



import sys

# global variable to indicate whether this is Python3 or not:
PY3 = sys.version_info[0] == 3


class FunctionMapper(object):
    """
    This class is used in Lexer, Parser, and Serializer to map IDs
    to functions"""
    def __init__(self, adict):
        self.adict = adict

    def __call__(self, *args):
        def wrap(func):
            for a in args:
                self.adict[a] = func
            return func
        return wrap


def hexString(aString):
    """
    convert a binary string in its hexadecimal representation,
    like '\x00\x01...'
    """
    if PY3:
        # in Py3 iterating over a byte-sequence directly provides the
        # numeric values of the bytes  ...
        return ''.join([r'\x%02x' % c for c in aString])
    else:
        # ... while in Py2 we need to use ord() to convert chars to
        # their numeric values:
        return ''.join([r'\x%02x' % ord(c) for c in aString])


def byteEncode(aString, encoding='utf-8'):
    # check for __name__ not to get faked by Python2.x!
    if PY3 and type(aString).__name__ != 'bytes':
        return bytes(aString, encoding=encoding)
    else:
        if type(aString).__name__.startswith('unicode'):
            return aString.encode('utf-8')
        else:
            return aString


def stringEncode(byteData, encoding='utf-8'):
    # check for __name__ not to get faked by Python2.x!
    if PY3 and type(byteData).__name__ == 'bytes':
        # got a real bytes object, must be python3 !
        return byteData.decode(encoding=encoding)
    else:
        # in py2.x there is no real byte-data, it is a string already
        return byteData


def padLen4(aString):
    """
    Calculate how many additional bytes a given string needs to have a length
    of a multiple of 4. A zero-length array is considered a multiple of 4.
    """
    l = len(aString)
    mod = divmod(l, 4)[1]
    return 4-mod if mod else 0


def string2bytesPad4(aString):
    """
    Return a given string converted into bytes, padded with zeros at the end
    to make its length be a multiple of 4.
    A zero-length string is considered a multiple of 4.
    """
    byteString = byteEncode(aString) + b'\0'
    return byteString + padLen4(byteString) * b'\0'



"""pyRserve package"""
import warnings
from .version import __version__

# Show all deprecated warning only once:
warnings.filterwarnings('once', category=DeprecationWarning)
del warnings

from .rconn import connect
from .taggedContainers import TaggedList, TaggedArray, AttrArray



"""
types module for pyRserve
"""
import numpy
from pyRserve.misc import PY3

# some general constants:
SOCKET_BLOCK_SIZE = 4096
MAX_INT32 = 2**31 - 1
MIN_INT32 = -MAX_INT32

# Rserve constants and mappings ###############################################

# Main Rserve header size [bytes]
RHEADER_SIZE = 16

# Header sizes (in SEXPR) without and with XT_LARGE or DT_LARGE flag [bytes]
SMALL_DATA_HEADER_SIZE = 4
LARGE_DATA_HEADER_SIZE = 8


CMD_RESP = 0x10000            # all responses have this flag set

RESP_OK  = CMD_RESP | 0x0001  # command succeeded; returned parameters depend
                              #   on the command issued
RESP_ERR = CMD_RESP | 0x0002  # command failed, check stats code


CMD_OOB         = 0x20000     # out-of-band data - i.e. unsolicited messages

OOB_SEND        = CMD_OOB | 0x1000 # OOB send - unsolicited SEXP sent from the
                                   # R instance to the client. 12 LSB are
                                   # reserved for application-specific code
OOB_MSG         = CMD_OOB | 0x2000 # OOB message - unsolicited message sent
                                   # from the R instance to the client
                                   # requiring a response. 12 LSB are reserved
                                   # for application-specific code
OOB_STREAM_READ = CMD_OOB | 0x4000 # OOB stream read request - server requests
                                   # streaming data from the client (typically
                                   # streaming input for computation)

###############################################################################
# Error codes

ERR_auth_failed      = 0x41  # auth.failed or auth.requested but no
                             #   login came. in case of authentification
                             #   failure due to name/pwd mismatch,
                             #   server may send CMD_accessDenied instead

ERR_conn_broken      = 0x42  # connection closed or broken packet killed it
ERR_inv_cmd          = 0x43  # unsupported/invalid command
ERR_inv_par          = 0x44  # some parameters are invalid
ERR_Rerror           = 0x45  # R-error occured, usually followed by connection
                             #   shutdown
ERR_IOerror          = 0x46  # I/O error
ERR_notOpen          = 0x47  # attempt to perform fileRead/Write on closed file
ERR_accessDenied     = 0x48  # this answer is also valid on
                             #   CMD_login; otherwise it's sent
                             #   if the server deosn;t allow the user
                             #   to issue the specified command.
                             #   (e.g. some server admins may block
ERR_unsupportedCmd   = 0x49  # unsupported command
ERR_unknownCmd       = 0x4a  # unknown command - the difference
                             #   between unsupported and unknown is that
                             #   unsupported commands are known to the
                             #   server but for some reasons (e.g.
                             #   platform dependent) it's not supported.
                             #   unknown commands are simply not recognized
                             #   by the server at all.

# The following ERR_.. exist since 1.23/0.1-6
ERR_data_overflow    = 0x4b  # incoming packet is too big.
                             #   currently there is a limit as of the
                             #   size of an incoming packet.
ERR_object_too_big   = 0x4c  # the requested object is too big
                             #   to be transported in that way.
                             #   If received after CMD_eval then
                             #   the evaluation itself was successful.
                             #   optional parameter is the size of the object

# since 1.29/0.1-9
ERR_out_of_mem       = 0x4d  # out of memory. the connection is usually
                             #  closed after this error was sent
# since 0.6-0
ERR_ctrl_closed      = 0x4e  # control pipe to the master process is closed
                             #  or broken

# since 0.4-0
ERR_session_busy     = 0x50  # session is still busy
ERR_detach_failed    = 0x51  # unable to detach seesion (cannot determine
                             #  peer IP or problems creating a listening socket
                             #  for resume)

# pack all error codes with their names into a dictionary:
ERRORS = dict([(errCode, err_name) for (err_name, errCode) in locals().items()
               if err_name.startswith('ERR_')])


###############################################################################
# Available commands

CMD_login        = 0x001    # "name\npwd" : -
CMD_voidEval     = 0x002    # string : -
CMD_eval         = 0x003    # string : encoded SEXP
CMD_shutdown     = 0x004    # [admin-pwd] : -

# file I/O routines. server may answe
CMD_openFile     = 0x010    # fn : -
CMD_createFile   = 0x011    # fn : -
CMD_closeFile    = 0x012    # - : -
CMD_readFile     = 0x013    # [int size] : data... ; if size not present,
                            #      server is free to choose any value - usually
                            #      it uses the size of its static buffer
CMD_writeFile    = 0x014    # data : -
CMD_removeFile   = 0x015    # fn : -

# object manipulation
CMD_setSEXP      = 0x020    # string(name), REXP : -
CMD_assignSEXP   = 0x021    # string(name), REXP : - ; same as setSEXP
                            #    except that the name is parsed

# session management (since 0.4-0)
CMD_detachSession    = 0x030  # : session key
CMD_detachedVoidEval = 0x031  # string : session key; doesn't
CMD_attachSession    = 0x032  # session key : -

# control commands (since 0.6-0) - passed on to the master process
# Note: currently all control commands are asychronous, i.e. RESP_OK
#   indicates that the command was enqueued in the master pipe, but there
#  is no guarantee that it will be processed. Moreover non-forked
#   connections (e.g. the default debug setup) don't process any
#   control commands until the current client connection is closed so
#   the connection issuing the control command will never see its result.

CMD_ctrl            = 0x40  # -- not a command - just a constant --
CMD_ctrlEval        = 0x42  # string : -
CMD_ctrlSource      = 0x45  # string : -
CMD_ctrlShutdown    = 0x44  # - : -

# 'internal' commands (since 0.1-9)
CMD_setBufferSize   = 0x081   # [int sendBufSize]
                              #     this commad allow clients to request
                              #     bigger buffer sizes if large data is to be
                              #     transported from Rserve to the client.
                              #     (incoming buffer is resized automatically)

CMD_setEncoding     = 0x082   # string (one of "native","latin1","utf8")

# special commands - the payload of packages with this mask does not contain
# defined parameters

CMD_SPECIAL_MASK    = 0xf0

CMD_serEval         = 0xf5     # serialized eval - the packets are raw
                               #   serialized data without data header
CMD_serAssign       = 0xf6     # serialized assign - serialized list with
                               #   [[1]]=name, [[2]]=value
CMD_serEEval        = 0xf7     # serialized expression eval - like serEval with
                               #   one additional evaluation round


###############################################################################
# Data types for the transport protocol (QAP1) do NOT confuse with any
# XT_.. values.

DT_INT           = 0x01  # int
DT_CHAR          = 0x02  # char
DT_DOUBLE        = 0x03  # double
DT_STRING        = 0x04  # 0 terminted string
DT_BYTESTREAM    = 0x05  # stream of bytes (unlike DT_STRING may contain 0)
DT_SEXP          = 0x0A  # encoded SEXP

DT_ARRAY         = 0x0B  # array of objects (i.e. first 4 bytes specify how
                         #   many subsequent objects are part of the array;
                         #   0 is legitimate)
DT_LARGE         = 0x40  # new in 0102: if this flag is set then the length of
                         #   the object is coded as 56-bit integer enlarging
                         # the header by 4 bytes

###############################################################################
# XpressionTypes

#   REXP - R expressions are packed in the same way as command parameters
#   transport format of the encoded Xpressions:
#   [0] int type/len (1 byte type, 3 bytes len - same as SET_PAR)
#   [4] REXP attr (if bit 8 in type is set)
#   [4/8] data ..

XT_NULL          =  0x00  # P  data: [0]
XT_INT           =  0x01  # -  data: [4]int
XT_DOUBLE        =  0x02  # -  data: [8]double
XT_STR           =  0x03  # P  data: [n]char null-term. strg.
XT_LANG          =  0x04  # -  data: same as XT_LIST
XT_SYM           =  0x05  # -  data: [n]char symbol name
XT_BOOL          =  0x06  # -  data: [1]byte boolean (1=TRUE, 0=FALSE, 2=NA)

XT_S4            =  0x07  # P  data: [0]

XT_BYTE          =  0x08  # extension for pyRserve
XT_INT3          =  0x09  # extension for pyRserve, a 3-byte integer as used
                          #   in REXP
XT_INT7          =  0x0A  # extension for pyRserve, a 7-byte integer as used
                          #   in REXP

XT_VECTOR        =  0x10  # 16dec: P  data: [?]REXP,REXP,..
XT_LIST          =  0x11  # 17dec: -  X head, X vals, X tag (since 0.1-5)
XT_CLOS          =  0x12  # 18dec: P  X formals, X body  (closure; since 0.1-5)
XT_SYMNAME       =  0x13  # 19dec: s  same as XT_STR (since 0.5)
XT_LIST_NOTAG    =  0x14  # 20dec: s  same as XT_VECTOR (since 0.5)
XT_LIST_TAG      =  0x15  # 21dec: P  X tag, X val, Y tag, Y val,  (since 0.5)
XT_LANG_NOTAG    =  0x16  # 22dec: s  same as XT_LIST_NOTAG (since 0.5)
XT_LANG_TAG      =  0x17  # 23dec: s  same as XT_LIST_TAG (since 0.5)
XT_VECTOR_EXP    =  0x1a  # 26dec: s  same as XT_VECTOR (since 0.5)
XT_VECTOR_STR    =  0x1b  # 27dec: -  same as XT_VECTOR (since 0.5 but unused,
                          #           use XT_ARRAY_STR instead)

XT_ARRAY_INT     =  0x20  # 32dec: P  data: [n*4]int,int,..
XT_ARRAY_DOUBLE  =  0x21  # 33dec: P  data: [n*8]double,double,..
XT_ARRAY_STR     =  0x22  # 34dec: P  data: string,string,.. (
                          #           string=byte,byte,...,0) padded with '\01'
XT_ARRAY_BOOL_UA =  0x23  # 35dec: -  data: [n]byte,byte,..
                          #           (unaligned! NOT supported anymore)
XT_ARRAY_BOOL    =  0x24  # 36dec: P  data: int(n),byte,byte,...
XT_RAW           =  0x25  # 37dec: P  data: int(n),byte,byte,...
XT_ARRAY_CPLX    =  0x26  # 38dec: P  data: [n*16]double,double,...
                          #           (Re,Im,Re,Im,...)

XT_UNKNOWN       =  0x30  # 48dec: P  data: [4]int - SEXP
#                                     type (as from TYPEOF(x))
#                      |
#                      +--- interesting flags for client implementations:
#                           P = primary type
#                           s = secondary type - its decoding is identical to
#                               a primary type and thus the client doesn't need
#                                to decode it separately.
#                           - = deprecated/removed. if a client doesn't need to
#                               support old Rserve versions, those can be
#                               safely skipped.
#  Total primary: 4 trivial types (NULL, STR, S4, UNKNOWN) + 6 array types +
#                 3 recursive types


XT_LARGE         =  0x40  # 64dec: new in 0102: if this flag is set then the
                          #   length of the object is coded as 56-bit integer
                          #   enlarging the header by 4 bytes
XT_HAS_ATTR      =  0x80  # 128dec: flag; if set, the following REXP is the
                          #   attribute the use of attributes and vectors
                          #   results in recursive storage of REXPs

# Build up a dictionary that translates all codes for XT_* and DT_* constants
# into their names:

XTs = dict([(rTypeCode, xt_name) for (xt_name, rTypeCode) in locals().items()
            if xt_name.startswith('XT_')])
DTs = dict([(rTypeCode, dt_name) for (dt_name, rTypeCode) in locals().items()
            if dt_name.startswith('DT_')])


BOOL_TRUE = 1
BOOL_FALSE = 0
BOOL_NA = 2

VALID_R_TYPES = [
    DT_SEXP, XT_BOOL, XT_INT, XT_DOUBLE, XT_STR, XT_SYMNAME, XT_VECTOR,
    XT_LIST_TAG, XT_LANG_TAG, XT_LIST_NOTAG, XT_LANG_NOTAG, XT_CLOS,
    XT_ARRAY_BOOL, XT_ARRAY_INT, XT_ARRAY_DOUBLE, XT_ARRAY_CPLX, XT_ARRAY_STR,
    XT_VECTOR_EXP, XT_NULL, XT_UNKNOWN, XT_RAW, XT_S4
]

STRING_TYPES = [str, numpy.string_, numpy.str_]
if not PY3:
    STRING_TYPES.append(unicode)

###############################################################################
# Mapping btw. numpy and R data types, in both directions

# map r-types and some python types to typecodes used in the 'struct' module
structMap = {
    XT_BOOL:          'b',
    numpy.bool:       'b',
    numpy.bool_:      'b',
    XT_BYTE:          'B',
    XT_INT:           'i',
    int:              'i',
    numpy.int32:      'i',
    XT_INT3:          'i',
    XT_INT7:          'q',     # 64 bit integer
    XT_DOUBLE:        'd',     # double (float64)
    float:            'd',
    numpy.double:     'd',
    complex:          'd',
    numpy.complex:    'd',
    numpy.complex128: 'd',
    }

# mapping to determine overall type of message.
DT_Map = {
    str:    DT_STRING,
    int:    DT_INT,
    float:  DT_DOUBLE,
}


numpyMap = {
    XT_ARRAY_BOOL:     numpy.bool_,
    #   XT_BYTE:           numpy.byte,
    XT_ARRAY_INT:      numpy.int32,
    XT_ARRAY_DOUBLE:   numpy.double,     # double float64
    XT_ARRAY_CPLX:     numpy.complex,
    XT_ARRAY_STR:      numpy.string_,
    }

# also add the inverse mapping to it:
for k, v in list(numpyMap.items()):
    numpyMap[v] = k

# some manual additions for numpy variants:
numpyMap[numpy.complex128] = XT_ARRAY_CPLX
numpyMap[numpy.int32]      = XT_ARRAY_INT
numpyMap[numpy.int64]      = XT_ARRAY_INT
numpyMap[numpy.long]       = XT_ARRAY_INT
numpyMap[numpy.str_]       = XT_ARRAY_STR
numpyMap[numpy.unicode_]   = XT_ARRAY_STR


atom2ArrMap = {
    # map atomic python objects to their array counterparts in R
    int:               XT_ARRAY_INT,
    numpy.int32:       XT_ARRAY_INT,
    float:             XT_ARRAY_DOUBLE,
    numpy.double:      XT_ARRAY_DOUBLE,
    complex:           XT_ARRAY_CPLX,
    numpy.complex:     XT_ARRAY_CPLX,
    numpy.complex128:  XT_ARRAY_CPLX,
    str:               XT_ARRAY_STR,
    numpy.str_:        XT_ARRAY_STR,
    numpy.string_:     XT_ARRAY_STR,
    numpy.unicode_:    XT_ARRAY_STR,
    bool:              XT_ARRAY_BOOL,
    numpy.bool:        XT_ARRAY_BOOL,
    numpy.bool_:       XT_ARRAY_BOOL,
    }


###############################################################################
# Some custom exceptions used in the pyRserve package


class RserveError(Exception):
    # base class for all errors for this package
    pass


class RParserError(RserveError):
    pass



# -*- coding: utf-8 -*-
"""
Module providing functionality to connect to a running Rserve instance
"""
import socket
import time
import pydoc
###
from . import rtypes
from .rexceptions import RConnectionRefused, REvalError, PyRserveClosed
from .rserializer import rEval, rAssign, rSerializeResponse, rShutdown
from .rparser import rparse, OOBMessage
from .misc import hexString

RSERVEPORT = 6311
DEBUG = False


def _defaultOOBCallback(data, code=0):
    return None


class OOBCallback(object):
    """Sets up conn with a new callback when entering the `with` block and
    restores the old one when exiting
    """
    def __init__(self, conn, callback):
        self.conn = conn
        self.callback = callback

    def __enter__(self):
        self.old_callback = self.conn.oobCallback
        self.conn.oobCallback = self.callback
        return self.conn

    def __exit__(self, exc_type, exc_value, traceback):
        self.conn.oobCallback = self.old_callback


def connect(host='', port=RSERVEPORT, atomicArray=False, defaultVoid=False,
            oobCallback=_defaultOOBCallback):
    """Open a connection to an Rserve instance
    Params:
    - host: provide hostname where Rserve runs, or leave as empty string to
            connect to localhost
    - port: Rserve port number, defaults to 6311
    - atomicArray:
            If True: when a result from an Rserve call is an array with
            a single element that single element
            is returned. Otherwise the array is returned unmodified.
            Default: True
    - arrayOrder:
            The order in which data in multi-dimensional arrays is returned.
            Provide 'C' for c-order, F for fortran. Default: 'C'
    - defaultVoid:
            If True then calls to conn.r('..') don't return a result by default
    - oobCallback:
            Callback to be executed when self.oobSend/oobMessage is called from
            R. The callback receives the submitted data and a user code as
            parameters. If self.oobMessage was used, the result value of the
            callback is sent back to R.
            Default: lambda data, code=0: None (oobMessage will return NULL)
    """
    if host in (None, ''):
        # On Win32 it seems that passing an empty string as 'localhost' does
        # not work. So just to be sure provide the full local hostname if None
        # or '' were passed.
        host = 'localhost'
    assert port is not None, 'port number must be given'
    return RConnector(host, port, atomicArray, defaultVoid, oobCallback)


def checkIfClosed(func):
    def decoCheckIfClosed(self, *args, **kw):
        if self.isClosed:
            raise PyRserveClosed('Connection to Rserve already closed')
        try:
            return func(self, *args, **kw)
        except socket.error as msg:
            if msg.strerror in ['Connection reset by peer', 'Broken pipe']:
                # seems like the connection to Rserve has died, so mark
                # the connection as closed
                self.close()
                raise PyRserveClosed('Connection to Rserve already closed')
            else:
                raise
    return decoCheckIfClosed


class RConnector(object):
    """Provide a network connector to an Rserve process"""
    def __init__(self, host, port, atomicArray, defaultVoid,
                 oobCallback=_defaultOOBCallback):
        self.sock = None
        self.__closed = True
        self.host = host
        self.port = port
        self.atomicArray = atomicArray
        self.defaultVoid = defaultVoid
        self.oobCallback = oobCallback
        self.r = RNameSpace(self)
        self.ref = RNameSpaceReference(self)
        self.connect()

    def __repr__(self):
        txt = 'Closed handle' if self.isClosed else 'Handle'
        return '<%s to Rserve on %s:%s>' % \
               (txt, self.host or 'localhost', self.port)

    @property
    def isClosed(self):
        return self.__closed

    def connect(self):
        self.sock = socket.socket()
        try:
            self.sock.connect((self.host, self.port))
        except socket.error:
            raise RConnectionRefused('Connection denied, server not reachable '
                                     'or not accepting connections')
        time.sleep(0.2)
        hdr = self.sock.recv(1024)
        self.__closed = False
        if DEBUG:
            print('received hdr %s from rserve' % hdr)
        # make sure we are really connected with rserv
        assert hdr.startswith(b'Rsrv01'), \
            'Protocol error with Rserv, obtained invalid header string'
        # TODO: possibly also do version checking here to make sure we
        #       understand the protocol...

    @checkIfClosed
    def close(self):
        """Close network connection to rserve"""
        self.sock.close()
        self.__closed = True

    @checkIfClosed
    def shutdown(self):
        rShutdown(fp=self.sock)
        self.close()

    def _reval(self, aString, void):
        rEval(aString, fp=self.sock, void=void)

    def _rrespond(self, aObj):
        rSerializeResponse(aObj, fp=self.sock)

    @checkIfClosed
    def eval(self, aString, atomicArray=None, void=False):
        """
        Evaluate a string expression through Rserve and return the result
        transformed into python objects
        """
        if not type(aString in rtypes.STRING_TYPES):
            raise TypeError('Only string evaluation is allowed')
        self._reval(aString, void)
        if DEBUG:
            # Read entire data into memory en bloque, it's easier to debug
            src = self._receive()
            print('Raw response: %s' % hexString(src))
        else:
            src = self.sock

        if atomicArray is None:
            # if not specified, use the global default:
            atomicArray = self.atomicArray

        try:
            message = rparse(src, atomicArray=atomicArray)
            # Before the result is returned, 0-∞ OOB messages may be sent
            while isinstance(message, OOBMessage):
                if DEBUG:
                    print('OOB Message received:', message)
                ret = self.oobCallback(message.data, message.userCode)
                if message.type == rtypes.OOB_MSG:
                    self._rrespond(ret)

                if isinstance(src, (str, bytes)):
                    # This is no stream, so we have to cut off data
                    src = src[len(message):]

                message = rparse(src, atomicArray=atomicArray)
            return message
        except REvalError:
            # R has reported an evaluation error, so let's obtain a descriptive
            # explanation about why the error has occurred. R allows to
            # retrieve the error message of the last exception via a built-in
            # function called 'geterrmessage()'.
            errorMsg = self.eval('geterrmessage()').strip()
            raise REvalError(errorMsg)

    @checkIfClosed
    def voidEval(self, aString):
        """
        Evaluate a string expression through Rserve without returning
        any result data
        """
        self.eval(aString, void=True)

    @checkIfClosed
    def _receive(self):
        """Receive the result from a previous call to rserve."""
        raw = self.sock.recv(rtypes.SOCKET_BLOCK_SIZE)
        d = [raw]
        while len(raw) == rtypes.SOCKET_BLOCK_SIZE:
            raw = self.sock.recv(rtypes.SOCKET_BLOCK_SIZE)
            d.append(raw)
        return ''.join(d)

#    @checkIfClosed
#    def _raw(self, *args, **kw):
#        self.send(*args)
#        return self.receive()

    @checkIfClosed
    def setRexp(self, name, o):
        """
        Convert a python object into an RExp and bind it to a variable
        called "name" in the R namespace
        """
        rAssign(name, o, self.sock)
        # Rserv sends an emtpy confirmation message, or error message in case
        # of an error. rparse() will raise an Exception in the latter case.
        rparse(self.sock, atomicArray=self.atomicArray)

    @checkIfClosed
    def getRexp(self, name):
        """Retrieve a Rexp stored in a variable called 'name'"""
        return self.eval(name)

    @checkIfClosed
    def callFunc(self, name, *args, **kw):
        """
        @brief  make a call to a function "name" through Rserve
        @detail positional and keyword arguments are first stored as local
                variables in the R namespace and then delivered to the
                function.
        @result Whatever the result of the called function is.
        """
        if name == 'rm':
            # SPECIAL HANDLING FOR "rm()":
            # Calling "rm" with real values instead of reference to values
            # works, however it doesn't produce the desired effect (it only
            # removes temporaily created variables). To avoid confusion for
            # the users a check is applied here to make sure that "args" only
            # contains variable or function references (proxies) and NOT
            # values!
            assert [x for x in args if not isinstance(x, RBaseProxy)] == (),\
                'Only references to variables or functions allowed for "rm()"'

        argNames = []
        for idx, arg in enumerate(args):
            if isinstance(arg, RBaseProxy):
                argName = arg.__name__
            else:
                # a real python value is passed. Set a value of an artificial
                # variable on the R side, memorize its name for making the
                # actual call to the function below
                argName = 'arg_%d_' % idx
                self.setRexp(argName, arg)
            argNames.append(argName)
        for key, value in kw.items():
            if isinstance(value, RBaseProxy):
                argName = value.__name__
            else:
                argName = 'kwarg_%s_' % key
                self.setRexp(argName, value)
            argNames.append('%s=%s' % (key, argName))
        return self.eval(name+'(%s)' % ', '.join(argNames))

    @checkIfClosed
    def assign(self, aDict):
        """Assign all items of the dictionary to the default R namespace"""
        for k, v in aDict.items():
            self.setRexp(k, v)

    @checkIfClosed
    def isFunction(self, name):
        """Check whether given name references an existing function in R"""
        return self.eval('is.function(%s)' % name)


class RNameSpace(object):
    """
    An instance of this class serves as access point to the default namesspace
    of an Rserve connection
    """
    def __init__(self, rconn):
        self.__dict__['_rconn'] = rconn

    def __setattr__(self, name, o):
        """Assign an rExpr to a variable called 'name'"""
        self._rconn.setRexp(name, o)

    def __getattr__(self, name):
        """
        Retrieve either Rexp stored in a variable called "name" or make call
        to function called 'name'
        """
        realname = name[1:] if name.startswith('_') else name
        try:
            isFunction = self._rconn.isFunction(realname)
        except:
            # an error is only raised if neither such a function or variable
            # exists at all!
            raise NameError('no such variable or function "%s" '
                            'defined in Rserve' % realname)
        if isFunction:
            return RFuncProxy(realname, self._rconn)
        elif name.startswith('_'):
            return RVarProxy(realname, self._rconn)
        else:
            return self._rconn.getRexp(name)

    def __call__(self, aString, atomicArray=None, void=None):
        if void is None:
            void = self._rconn.defaultVoid
        return self._rconn.eval(aString, atomicArray=atomicArray, void=void)


class RNameSpaceReference(object):
    """
    Provide reference to R objects (a proxy), NOT directly to their values
    """
    def __init__(self, rconn):
        self.__dict__['_rconn'] = rconn

    def __getattr__(self, name):
        """Return either a reference proxy to a variable to to a function"""
        try:
            isFunction = self._rconn.isFunction(name)
        except:
            # an error is only raised if neither such a function or variable
            # exists at all!
            raise NameError('no such variable or function "%s" '
                            'defined in Rserve' % name)
        if isFunction:
            return RFuncProxy(name, self._rconn)
        else:
            return RVarProxy(name, self._rconn)


class RBaseProxy(object):
    """
    Proxy for a reference to a variable or function in R.
    Do not use this directly, only its subclasses.
    """
    def __init__(self, name, rconn):
        self.__name__ = name
        self._rconn = rconn


class RVarProxy(RBaseProxy):
    """Proxy for a reference to a variable in R"""
    def __repr__(self):
        return '<RVarProxy to variable "%s">' % self.__name__

    def value(self):
        return self._rconn.getRexp(self.__name__)


class RFuncProxy(RBaseProxy):
    """Proxy for function calls to Rserve"""
    def __repr__(self):
        return '<RFuncProxy to function "%s">' % self.__name__

    def __call__(self, *args, **kw):
        return self._rconn.callFunc(self.__name__, *args, **kw)

    # command to send to R in order to get the help for a function in text
    # format:
    R_HELP = "capture.output(tools:::Rd2txt(utils:::.getHelpFile(help(%s))))"

    @property
    def __doc__(self):
        """
        There are different ways to get the help message from R:
        # The the package db file:
        pkgRdDB = tools:::fetchRdDB(file.path(find.package('base'),
                                              'help', 'base'))
        # show all available topics in the help package:
        names(pkgRdDB)
        # convert the 'lapply' help message to text (from the base package):
        tools::Rd2txt(pkgRdDB[['lapply']])
        # capture this output into a variable:
        a <- capture.output(tools::Rd2txt(pkgRdDB[['lapply']]))
        Disadvantage: One needs to know the package beforehand.

        Better:
        Everything in one line and better (doesn't need to know the pkg):
        a <- capture.output(tools:::Rd2txt(utils:::.getHelpFile(help(sapply))))
        """
        try:
            d = self._rconn.eval(self.R_HELP % self.__name__)
        except REvalError:
            # probably no help available, unfortunately there is no specific
            # code for this...
            return None
        # Join the list of strings:
        helpstring = '\n'.join(d)
        # remove some obscure characters:
        # helpstring = helpstring.replace('_\x08', '')
        return helpstring

    def help(self):
        """Directly page the help message to the terminal (e.g. via less)"""
        pydoc.pager(self.__doc__)

    def __getattr__(self, name):
        """Allow for nested name space calls, e.g. 't.test' """
        if name == '__name__':
            # this is useful for py.test which does some code inspection
            # during runtime
            return self.__name__

        concatName = "%s.%s" % (self.__name__, name)
        try:
            self._rconn.isFunction(concatName)
        except:
            # an error is only raised if neither such a function or variable
            # exists at all!
            raise NameError('no such variable or function "%s" '
                            'defined in R' % concatName)
        return RFuncProxy(concatName, self._rconn)


def _test_main():
    import os
    import readline
    import atexit
    # Setup history and readline facility for remote q:
    histfile = os.path.join(os.environ['HOME'], '.pyhistory')
    try:
        readline.read_history_file(histfile)
    except IOError:
        pass
    atexit.register(readline.write_history_file, histfile)

    conn = connect()
    print('"conn" is your handle to rserve. Type e.g. "conn(\'1\')" '
          'for string evaluation.')
    # r('x<-1:20; y<-x*2; lm(y~x)')
    sc = open('../testData/test-script.R').read()
    v = conn.r(sc)
    open('r-test-png.png', 'w').write(v[3])
    conn.r.v = 'abc'
    conn.r('func0 <- function() { 3 }')
    conn.r('func1 <- function(a1) { a1 }')
    conn.r('func2 <- function(a1, a2) { list(a1, a2) }')
    conn.r('funcKW <- function(a1=1, a2=4) { list(a1, a2) }')
    conn.r('squared<-function(t) t^2')


if __name__ == '__main__':
    _test_main()
